/*
 * Copyright (c) 2003 by Quinton Dolan <q@onthenet.com.au>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: nvidia_api.c,v 1.2 2005/06/21 02:37:02 wsm Exp $
 */
/*
 * NVIDIA nForce MCP Networking Adapter driver
 *
 * This is a port of the NVIDIA MCP Linux ethernet driver distributed by NVIDIA
 * through their web site.
 *
 * All mainstream nForce and nForce2 motherboards are supported. This module
 * is as stable, sometimes more stable, than the linux version. (Recent
 * Linux stability issues seem to be related to some issues with newer
 * distributions using GCC 3.x, however this don't appear to effect FreeBSD
 * 5.x).
 *
 * In accordance with the NVIDIA distribution license it is necessary to
 * link this module against the nvlibnet.o binary object included in the
 * Linux driver source distribution. The binary component is not modified in
 * any way and is simply linked against a FreeBSD equivalent of the nvnet.c
 * linux kernel module "wrapper".
 *
 * The Linux driver uses a common code API that is shared between Win32 and
 * i386 Linux. This abstracts the low level driver functions and uses
 * callbacks and hooks to access the underlying hardware device. By using
 * this same API in a FreeBSD kernel module it is possible to support the
 * hardware without breaching the Linux source distributions licensing
 * requirements, or obtaining the hardware programming specifications.
 *
 * Although not conventional, it works, and given the relatively small
 * amount of hardware centric code, it's hopefully no more buggy than its
 * linux counterpart.
 *
 * NVIDIA now suppport the nForce3 AMD64 platform, however I have been
 * unable to access such a system to verify support. However, the code is
 * reported to work with little modification when compiled with the AMD64
 * version of the NVIDIA Linux library. All that should be necessary to make
 * the driver work is to link it directly into the kernel, instead of as a
 * module, and apply the docs/amd64.diff patch in this source distribution to
 * the NVIDIA Linux driver source.
 *
 * This driver should work on all versions of FreeBSD since 4.9/5.1 as well
 * as recent versions of DragonFly.
 *
 * Written by Quinton Dolan <q@onthenet.com.au>
 * Portions based on existing FreeBSD network drivers.
 * NVIDIA API usage derived from distributed NVIDIA NVNET driver source files.
 *
 */

/* Ported to NetBSD by William S. Morgart <wsm@morgart.com> from FreeBSD
 * - April 2005
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <sys/ksyms.h>
#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>
#include <sys/lkm.h>
#include <sys/sockio.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_ether.h>
#include <net/if_media.h>
#include <dev/mii/miivar.h>

#include "nvidia_api.h"
#include "nvnet.h"

/* nvidia interface api */

static NV_SINT32 nv_osalloc(PNV_VOID, PMEMORY_BLOCK);
static NV_SINT32 nv_osfree(PNV_VOID, PMEMORY_BLOCK);
static NV_SINT32 nv_osallocex(PNV_VOID, PMEMORY_BLOCKEX);
static NV_SINT32 nv_osfreeex(PNV_VOID, PMEMORY_BLOCKEX);
static NV_SINT32 nv_osclear(PNV_VOID, PNV_VOID, NV_SINT32);
static NV_SINT32 nv_osdelay(PNV_VOID, NV_UINT32);
static NV_SINT32 nv_osallocrxbuf(PNV_VOID, PMEMORY_BLOCK, PNV_VOID *);
static NV_SINT32 nv_osfreerxbuf(PNV_VOID, PMEMORY_BLOCK, PNV_VOID);
static NV_SINT32 nv_ospackettx(PNV_VOID, PNV_VOID, NV_UINT32);
static NV_SINT32 nv_ospacketrx(PNV_VOID, PNV_VOID, NV_UINT32, NV_UINT8 *, NV_UINT8);
static NV_SINT32 nv_oslinkchg(PNV_VOID, NV_SINT32);
static NV_SINT32 nv_osalloctimer(PNV_VOID, PNV_VOID *);
static NV_SINT32 nv_osfreetimer(PNV_VOID, PNV_VOID);
static NV_SINT32 nv_osinittimer(PNV_VOID, PNV_VOID, PTIMER_FUNC, PNV_VOID);
static NV_SINT32 nv_ossettimer(PNV_VOID, PNV_VOID, NV_UINT32);
static NV_SINT32 nv_oscanceltimer(PNV_VOID, PNV_VOID);

static NV_SINT32 nv_ospreprocpkt(PNV_VOID, PNV_VOID, PNV_VOID *, NV_UINT8 *, NV_UINT8);
static PNV_VOID nv_ospreprocpktnopq(PNV_VOID, PNV_VOID);
static NV_SINT32 nv_osindicatepkt(PNV_VOID, PNV_VOID *, NV_UINT32);
static NV_SINT32 nv_oslockalloc(PNV_VOID, NV_SINT32, PNV_VOID *);
static NV_SINT32 nv_oslockacquire(PNV_VOID, NV_SINT32, PNV_VOID);
static NV_SINT32 nv_oslockrelease(PNV_VOID, NV_SINT32, PNV_VOID);
static PNV_VOID nv_osreturnbufvirt(PNV_VOID, PNV_VOID);

OS_API *
nvidia_api_init(struct nv_softc * sc)
{
	OS_API *osapi = &sc->osapi;
	osapi->pOSCX = sc;
	osapi->pfnAllocMemory = nv_osalloc;
	osapi->pfnFreeMemory = nv_osfree;
	osapi->pfnAllocMemoryEx = nv_osallocex;
	osapi->pfnFreeMemoryEx = nv_osfreeex;
	osapi->pfnClearMemory = nv_osclear;
	osapi->pfnStallExecution = nv_osdelay;
	osapi->pfnAllocReceiveBuffer = nv_osallocrxbuf;
	osapi->pfnFreeReceiveBuffer = nv_osfreerxbuf;
	osapi->pfnPacketWasSent = nv_ospackettx;
	osapi->pfnPacketWasReceived = nv_ospacketrx;
	osapi->pfnLinkStateHasChanged = nv_oslinkchg;
	osapi->pfnAllocTimer = nv_osalloctimer;
	osapi->pfnFreeTimer = nv_osfreetimer;
	osapi->pfnInitializeTimer = nv_osinittimer;
	osapi->pfnSetTimer = nv_ossettimer;
	osapi->pfnCancelTimer = nv_oscanceltimer;
	osapi->pfnPreprocessPacket = nv_ospreprocpkt;
	osapi->pfnPreprocessPacketNopq = nv_ospreprocpktnopq;
	osapi->pfnIndicatePackets = nv_osindicatepkt;
	osapi->pfnLockAlloc = nv_oslockalloc;
	osapi->pfnLockAcquire = nv_oslockacquire;
	osapi->pfnLockRelease = nv_oslockrelease;
	osapi->pfnReturnBufferVirtual = nv_osreturnbufvirt;
	return osapi;
}

int
nvidia_hardware_init(struct nv_softc * sc,
    int max_dpc_loop,
    int max_rx_size,
    int max_tx_size,
    int max_rx_pkt_buffer,
    int nv_pollinterval)
{
	ADAPTER_OPEN_PARAMS op;
	int error;

	/* Set NVIDIA API startup parameters */
	bzero(&op, sizeof(ADAPTER_OPEN_PARAMS));
	op.MaxDpcLoop = max_dpc_loop;
	op.MaxRxPkt = max_rx_size;
	op.MaxTxPkt = max_tx_size;
	op.SentPacketStatusSuccess = 1;
	op.SentPacketStatusFailure = 0;
	op.MaxRxPktToAccumulate = max_rx_pkt_buffer;
	op.ulPollInterval = nv_pollinterval;
	op.SetForcedModeEveryNthRxPacket = 0;
	op.SetForcedModeEveryNthTxPacket = 0;
	op.RxForcedInterrupt = 0;
	op.TxForcedInterrupt = 0;
	op.pOSApi = &sc->osapi;
	op.pvHardwareBaseAddress = bus_space_vaddr(sc->sc_st, sc->sc_sh);
	op.bASFEnabled = 0;
	op.ulDescriptorVersion = sc->hwmode;
	op.ulMaxPacketSize = sc->max_frame_size;
	op.DeviceId = sc->pci_device_id;
	error = ADAPTER_Open(&op, (void **) &(sc->hwapi), &sc->phyaddr);
	if (error) {
		aprint_error("failed to open NVIDIA Hardware API: 0x%x\n", error);
		return -1;
	}
	return 0;
}
/* Allocate DMA enabled general use memory for API */
static NV_SINT32
nv_osalloc(PNV_VOID ctx, PMEMORY_BLOCK mem)
{
	struct nv_softc *sc;
	bus_addr_t mem_physical;

	//DEBUGOUT(NV_DEBUG_API, "nv: nv_osalloc - %d\n", mem->uiLength);

	sc = (struct nv_softc *) ctx;

	mem->pLogical = (PVOID) malloc(mem->uiLength,
	    M_DEVBUF,
	    M_NOWAIT | M_ZERO);

	if (!mem->pLogical) {
		aprint_error("%s: memory allocation failed\n",
		    sc->sc_dev.dv_xname);
		return (0);
	}
	memset(mem->pLogical, 0, (ulong) mem->uiLength);
	mem_physical = vtophys(mem->pLogical);
	mem->pPhysical = (PVOID) mem_physical;

	//DEBUGOUT(NV_DEBUG_API, "nv: nv_osalloc 0x%x/0x%x - %d\n",
	    //(uint) mem->pLogical,
	    //(uint) mem->pPhysical, (uint) mem->uiLength);

	return (1);
}
/* Free allocated memory */
static NV_SINT32
nv_osfree(PNV_VOID ctx, PMEMORY_BLOCK mem)
{
	//DEBUGOUT(NV_DEBUG_API, "nv: nv_osfree - 0x%x - %d\n",
	    //(uint) mem->pLogical, (uint) mem->uiLength);

	free(mem->pLogical, M_DEVBUF);

	return (1);
}
/* Copied directly from nvnet.c */
static NV_SINT32
nv_osallocex(PNV_VOID ctx, PMEMORY_BLOCKEX mem_block_ex)
{
	MEMORY_BLOCK mem_block;

	//DEBUGOUT(NV_DEBUG_API, "nv: nv_osallocex\n");

	mem_block_ex->pLogical = NULL;
	mem_block_ex->uiLengthOrig = mem_block_ex->uiLength;

	if ((mem_block_ex->AllocFlags & ALLOC_MEMORY_ALIGNED) &&
	    (mem_block_ex->AlignmentSize > 1)) {
		//DEBUGOUT(NV_DEBUG_API, "     aligning on %d\n",
		    //mem_block_ex->AlignmentSize);
		mem_block_ex->uiLengthOrig += mem_block_ex->AlignmentSize;
	}
	mem_block.uiLength = mem_block_ex->uiLengthOrig;

	if (nv_osalloc(ctx, &mem_block) == 0) {
		return (0);
	}
	mem_block_ex->pLogicalOrig = mem_block.pLogical;
	mem_block_ex->pPhysicalOrigLow = (unsigned long) mem_block.pPhysical;
	mem_block_ex->pPhysicalOrigHigh = 0;

	mem_block_ex->pPhysical = mem_block.pPhysical;
	mem_block_ex->pLogical = mem_block.pLogical;

	if (mem_block_ex->uiLength != mem_block_ex->uiLengthOrig) {
		unsigned int offset;
		offset = mem_block_ex->pPhysicalOrigLow & (mem_block_ex->AlignmentSize - 1);

		if (offset) {
			mem_block_ex->pPhysical = (PVOID) ((ulong) mem_block_ex->pPhysical +
			    mem_block_ex->AlignmentSize - offset);
			mem_block_ex->pLogical = (PVOID) ((ulong) mem_block_ex->pLogical +
			    mem_block_ex->AlignmentSize - offset);
		}		/* if (offset) */
	}			/* if (mem_block_ex->uiLength !=
				 * mem_block_ex->uiLengthOrig) */
	return (1);
}
/* Copied directly from nvnet.c */
static NV_SINT32
nv_osfreeex(PNV_VOID ctx, PMEMORY_BLOCKEX mem_block_ex)
{
	MEMORY_BLOCK mem_block;

	//DEBUGOUT(NV_DEBUG_API, "nv: nv_osfreeex\n");

	mem_block.pLogical = mem_block_ex->pLogicalOrig;
	mem_block.pPhysical = (PVOID) ((ulong) mem_block_ex->pPhysicalOrigLow);
	mem_block.uiLength = mem_block_ex->uiLengthOrig;

	return (nv_osfree(ctx, &mem_block));
}
/* Clear memory region */
static NV_SINT32
nv_osclear(PNV_VOID ctx, PNV_VOID mem, NV_SINT32 length)
{
	//DEBUGOUT(NV_DEBUG_API, "nv: nv_osclear\n");
	memset(mem, 0, length);
	return (1);
}
/* Sleep for a tick */
static NV_SINT32
nv_osdelay(PNV_VOID ctx, NV_UINT32 usec)
{
	DELAY(usec);
	return (1);
}
/* Allocate memory for rx buffer */
static NV_SINT32
nv_osallocrxbuf(PNV_VOID ctx, PMEMORY_BLOCK mem, PNV_VOID * id)
{
	struct nv_softc *sc = ctx;
	struct nv_rx_desc *desc;
	struct nv_map_buffer *buf;
	int error;

	NV_LOCK(sc);

	DEBUGOUT(NV_DEBUG_API, "nv: nv_osallocrxbuf\n");

	if (sc->pending_rxs == RX_RING_SIZE) {
		aprint_error("%s: nv_osallocrxbuf - rx ring buffer is full\n",
		    sc->sc_dev.dv_xname);
		goto fail;
	}
	desc = sc->rx_desc + sc->cur_rx;
	buf = &desc->buf;

	if (buf->mbuf == NULL) {
		MGETHDR(buf->mbuf, M_DONTWAIT, MT_DATA);
		if (buf->mbuf == NULL)
			goto fail;
		MCLGET(buf->mbuf, M_DONTWAIT);
		if (buf->mbuf == NULL) {
			aprint_error("%s: nv_osallocrxbuf - failed to allocate memory\n",
			    sc->sc_dev.dv_xname);
			goto fail;
		}
		buf->mbuf->m_len = buf->mbuf->m_pkthdr.len = MCLBYTES;
		m_adj(buf->mbuf, ETHER_ALIGN);

		error = bus_dmamap_create(sc->sc_dmat,
		    buf->mbuf->m_len, 1,
		    buf->mbuf->m_len, 0,
		    BUS_DMA_NOWAIT,
		    &buf->map);
		if (error) {
			aprint_error("%s: nv_osallocrxbuf - couldn't create dma map\n",
			    sc->sc_dev.dv_xname);
			m_freem(buf->mbuf);
			buf->mbuf = NULL;
			goto fail;
		}
		error = bus_dmamap_load_mbuf(sc->sc_dmat,
		    buf->map,
		    buf->mbuf,
		    BUS_DMA_NOWAIT);
		if (error) {
			aprint_error("%s: nv_osallocrxbuf - failed to dma map mbuf\n",
			    sc->sc_dev.dv_xname);
			m_freem(buf->mbuf);
			buf->mbuf = NULL;
			goto fail;
		}
		bus_dmamap_sync(sc->sc_dmat, buf->map, 0, buf->mbuf->m_len, BUS_DMASYNC_PREREAD);
		desc->buflength = buf->mbuf->m_len;
		desc->vaddr = mtod(buf->mbuf, caddr_t);
		desc->paddr = vtophys(desc->vaddr);
	}
	sc->pending_rxs++;
	sc->cur_rx = (sc->cur_rx + 1) % RX_RING_SIZE;

	mem->pLogical = (void *) desc->vaddr;
	mem->pPhysical = (void *) desc->paddr;
	mem->uiLength = desc->buflength;
	DEBUGOUT(NV_DEBUG_API, "nv: nv_osallocrxbuf 0x%x/0x%x - %d\n",
	    (uint) mem->pLogical,
	    (uint) mem->pPhysical,
	    (uint) mem->uiLength);


	*id = (void *) desc;

	NV_UNLOCK(sc);
	return (1);

fail:
	NV_UNLOCK(sc);
	return (0);
}


/* Free the rx buffer */
static NV_SINT32
nv_osfreerxbuf(PNV_VOID ctx, PMEMORY_BLOCK mem, PNV_VOID id)
{
	struct nv_softc *sc = ctx;
	struct nv_rx_desc *desc;
	struct nv_map_buffer *buf;

	NV_LOCK(sc);

	DEBUGOUT(NV_DEBUG_API, "nv: nv_osfreerxbuf\n");

	desc = (struct nv_rx_desc *) id;
	buf = &desc->buf;

	if (buf->mbuf) {
		bus_dmamap_unload(sc->sc_dmat, buf->map);
		bus_dmamap_destroy(sc->sc_dmat, buf->map);
		m_freem(buf->mbuf);
	}
	sc->pending_rxs--;
	buf->mbuf = NULL;

	NV_UNLOCK(sc);

	return (1);
}
/* This gets called by the Nvidia API after our TX packet has been sent */
static NV_SINT32
nv_ospackettx(PNV_VOID ctx, PNV_VOID id, NV_UINT32 success)
{
	struct nv_softc *sc = ctx;
	struct nv_map_buffer *buf;
	struct nv_tx_desc *desc = (struct nv_tx_desc *) id;
	struct ifnet *ifp;

	NV_LOCK(sc);

	DEBUGOUT(NV_DEBUG_API, "nv: nv_ospackettx\n");

	ifp = &sc->sc_if;
	buf = &desc->buf;
	sc->pending_txs--;

	/* Unload and free mbuf cluster */
	if (buf->mbuf == NULL)
		goto fail;

	bus_dmamap_sync(sc->sc_dmat,
	    buf->map,
	    0,
	    buf->mbuf->m_len,
	    BUS_DMASYNC_POSTWRITE);
	bus_dmamap_unload(sc->sc_dmat, buf->map);
	m_freem(buf->mbuf);
	buf->mbuf = NULL;

	/* Send more packets if we have them */
	if (sc->pending_txs < TX_RING_SIZE)
		sc->sc_if.if_flags &= ~IFF_OACTIVE;

	if (ifp->if_snd.ifq_head != NULL && sc->pending_txs < TX_RING_SIZE)
		(*ifp->if_start) (ifp);

fail:
	NV_UNLOCK(sc);

	return (1);
}
/* This gets called by the Nvidia API when a new packet has been received */
/* XXX What is newbuf used for? XXX */
static NV_SINT32
nv_ospacketrx(PNV_VOID ctx, PNV_VOID data, NV_UINT32 success,
    NV_UINT8 * newbuf, NV_UINT8 priority)
{
	struct nv_softc *sc = ctx;
	struct ifnet *ifp;
	struct nv_rx_desc *desc;
	struct nv_map_buffer *buf;
	ADAPTER_READ_DATA *readdata;

	NV_LOCK(sc);

	DEBUGOUT(NV_DEBUG_API, "nv: nv_ospacketrx\n");

	ifp = &sc->sc_if;

	readdata = (ADAPTER_READ_DATA *) data;
	desc = readdata->pvID;
	buf = &desc->buf;
	bus_dmamap_sync(sc->sc_dmat,
	    buf->map,
	    0, buf->mbuf->m_len,
	    BUS_DMASYNC_POSTREAD);

	if (success) {
		/* Sync DMA bounce buffer. */
		bus_dmamap_sync(sc->sc_dmat,
		    buf->map,
		    0, buf->mbuf->m_len,
		    BUS_DMASYNC_POSTREAD);

		/* First mbuf in packet holds the ethernet and packet headers */
		buf->mbuf->m_pkthdr.rcvif = ifp;
		buf->mbuf->m_pkthdr.len = buf->mbuf->m_len = readdata->ulTotalLength;

		//bus_dmamap_unload(sc->sc_dmat, buf->map);
		bus_dmamap_destroy(sc->sc_dmat, buf->map);

		/* Pass packet up to any BPF listeners */
		/*   from Dave Huang <khm@azeotrope.org> */
		BPF_MTAP(ifp, buf->mbuf);

		/* Give mbuf to OS. */
		(*ifp->if_input) (ifp, buf->mbuf);
		if (readdata->ulFilterMatch & ADREADFL_MULTICAST_MATCH)
			ifp->if_imcasts++;

		/* Blat the mbuf pointer, kernel will free the mbuf cluster */
		buf->mbuf = NULL;
	} else {
		bus_dmamap_sync(sc->sc_dmat,
		    buf->map,
		    0, buf->mbuf->m_len,
		    BUS_DMASYNC_POSTREAD);
		bus_dmamap_unload(sc->sc_dmat, buf->map);
		m_freem(buf->mbuf);
		buf->mbuf = NULL;
	}

	sc->cur_rx = desc - sc->rx_desc;
	sc->pending_rxs--;

	NV_UNLOCK(sc);

	return (1);
}
/* This gets called by NVIDIA API when the PHY link state changes */
static NV_SINT32
nv_oslinkchg(PNV_VOID ctx, NV_SINT32 enabled)
{
	struct nv_softc *sc = (struct nv_softc *) ctx;
	struct ifnet *ifp;

	DEBUGOUT(NV_DEBUG_API, "nv: nv_oslinkchg\n");

	ifp = &sc->sc_if;

	if (enabled)
		ifp->if_flags |= IFF_UP;
	else
		ifp->if_flags &= ~IFF_UP;


	return (1);
}


/* Setup a watchdog timer */
static NV_SINT32
nv_osalloctimer(PNV_VOID ctx, PNV_VOID * timer)
{
	struct nv_softc *sc = (struct nv_softc *) ctx;

	//DEBUGOUT(NV_DEBUG_BROKEN, "nv: nv_osalloctimer\n");

	callout_init(&sc->ostimer);
	*timer = &sc->ostimer;

	return (1);
}
/* Free the timer */
static NV_SINT32
nv_osfreetimer(PNV_VOID ctx, PNV_VOID timer)
{
	//DEBUGOUT(NV_DEBUG_BROKEN, "nv: nv_osfreetimer\n");

	return (1);
}
/* Setup timer parameters */
static NV_SINT32
nv_osinittimer(PNV_VOID ctx, PNV_VOID timer, PTIMER_FUNC func, PNV_VOID parameters)
{
	struct nv_softc *sc = (struct nv_softc *) ctx;

	//DEBUGOUT(NV_DEBUG_BROKEN, "nv: nv_osinittimer\n");

	sc->ostimer_func = func;
	sc->ostimer_params = parameters;

	return (1);
}
/* Set the timer to go off */
static NV_SINT32
nv_ossettimer(PNV_VOID ctx, PNV_VOID timer, NV_UINT32 delay)
{
	struct nv_softc *sc = ctx;
	struct callout *c = (struct callout *) timer;
	//DEBUGOUT(NV_DEBUG_BROKEN, "nv: nv_ossettimer\n");

	callout_setfunc(c, sc->ostimer_func, sc->ostimer_params);
	callout_schedule(c, delay);

	return (1);
}
/* Cancel the timer */
static NV_SINT32
nv_oscanceltimer(PNV_VOID ctx, PNV_VOID timer)
{
	struct nv_softc *sc = ctx;

	//DEBUGOUT(NV_DEBUG_BROKEN, "nv: nv_oscanceltimer\n");

	callout_stop((struct callout *) timer);

	return (1);
}

static NV_SINT32
nv_ospreprocpkt(PNV_VOID ctx, PNV_VOID readdata, PNV_VOID * id, NV_UINT8 * newbuffer,
    NV_UINT8 priority)
{
	/* Not implemented */
	DEBUGOUT(NV_DEBUG_BROKEN, "nv: nv_ospreprocpkt\n");

	return (1);
}

static PNV_VOID
nv_ospreprocpktnopq(PNV_VOID ctx, PNV_VOID readdata)
{
	/* Not implemented */
	DEBUGOUT(NV_DEBUG_BROKEN, "nv: nv_ospreprocpkt\n");

	return (NULL);
}

static NV_SINT32
nv_osindicatepkt(PNV_VOID ctx, PNV_VOID * id, NV_UINT32 pktno)
{
	/* Not implemented */
	DEBUGOUT(NV_DEBUG_BROKEN, "nv: nv_osindicatepkt\n");

	return (1);
}
/* Allocate mutex context (already done in nv_attach) */
static NV_SINT32
nv_oslockalloc(PNV_VOID ctx, NV_SINT32 type, PNV_VOID * pLock)
{
	struct nv_softc *sc = (struct nv_softc *) ctx;

	*pLock = (void **) sc;

	return (1);
}
/* Obtain a spin lock */
static NV_SINT32
nv_oslockacquire(PNV_VOID ctx, NV_SINT32 type, PNV_VOID lock)
{
	struct nv_softc *sc = (struct nv_softc *) lock;

	sc->spl = splnet();
	return (1);
}
/* Release lock */
static NV_SINT32
nv_oslockrelease(PNV_VOID ctx, NV_SINT32 type, PNV_VOID lock)
{
	struct nv_softc *sc = (struct nv_softc *) lock;

	splx(sc->spl);

	return (1);
}
/* I have no idea what this is for */
static PNV_VOID
nv_osreturnbufvirt(PNV_VOID ctx, PNV_VOID readdata)
{
	/* Not implemented */
	DEBUGOUT(NV_DEBUG_LOCK, "nv: nv_osreturnbufvirt\n");
	panic("nv: nv_osreturnbufvirtual not implemented\n");

	return (NULL);
}
