/*
 * Copyright (c) 2003 by Quinton Dolan <q@onthenet.com.au>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS `AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: nvnet_drv.c,v 1.1.1.1 2005/06/21 02:18:41 wsm Exp $
 */
/*
 * NVIDIA nForce MCP Networking Adapter driver
 *
 * This is a port of the NVIDIA MCP Linux ethernet driver distributed by NVIDIA
 * through their web site.
 *
 * All mainstream nForce and nForce2 motherboards are supported. This module
 * is as stable, sometimes more stable, than the linux version. (Recent
 * Linux stability issues seem to be related to some issues with newer
 * distributions using GCC 3.x, however this don't appear to effect FreeBSD
 * 5.x).
 *
 * In accordance with the NVIDIA distribution license it is necessary to
 * link this module against the nvlibnet.o binary object included in the
 * Linux driver source distribution. The binary component is not modified in
 * any way and is simply linked against a FreeBSD equivalent of the nvnet.c
 * linux kernel module "wrapper".
 *
 * The Linux driver uses a common code API that is shared between Win32 and
 * i386 Linux. This abstracts the low level driver functions and uses
 * callbacks and hooks to access the underlying hardware device. By using
 * this same API in a FreeBSD kernel module it is possible to support the
 * hardware without breaching the Linux source distributions licensing
 * requirements, or obtaining the hardware programming specifications.
 *
 * Although not conventional, it works, and given the relatively small
 * amount of hardware centric code, it's hopefully no more buggy than its
 * linux counterpart.
 *
 * NVIDIA now suppport the nForce3 AMD64 platform, however I have been
 * unable to access such a system to verify support. However, the code is
 * reported to work with little modification when compiled with the AMD64
 * version of the NVIDIA Linux library. All that should be necessary to make
 * the driver work is to link it directly into the kernel, instead of as a
 * module, and apply the docs/amd64.diff patch in this source distribution to
 * the NVIDIA Linux driver source.
 *
 * This driver should work on all versions of FreeBSD since 4.9/5.1 as well
 * as recent versions of DragonFly.
 *
 * Written by Quinton Dolan <q@onthenet.com.au>
 * Portions based on existing FreeBSD network drivers.
 * NVIDIA API usage derived from distributed NVIDIA NVNET driver source files.
 *
 */

/* Ported to NetBSD by William S. Morgart <wsm@morgart.com> from FreeBSD
 * - April 2005
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <sys/ksyms.h>
#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>
#include <sys/lkm.h>
#include <sys/sockio.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_ether.h>
#include <net/if_media.h>

#define MII_VERBOSE 1

#include <dev/mii/miivar.h>

#include "nvnet.h"

static char *format_txdesc(struct nv_tx_desc * desc);

static int nv_match(struct device *, struct cfdata *, void *);
static void nv_attach(struct device *, struct device *, void *);
static int nv_detach(struct device *, int);

CFATTACH_DECL(nv, sizeof(struct nv_softc),
    nv_match, nv_attach, nv_detach, NULL);

static const struct nv_type {
	u_int32_t nv_prodid;
	const char *nv_name;
}       nv_devs[] = {
	{
		NFORCE_MCPNET1_DEVICEID,
		"NVIDIA nForce MCP Networking Adapter"
	},
	{
		NFORCE_MCPNET2_DEVICEID,
		"NVIDIA nForce MCP2 Networking Adapter"
	},
	{
		NFORCE_MCPNET3_DEVICEID,
		"NVIDIA nForce MCP3 Networking Adapter"
	},
	{
		NFORCE_MCPNET3_DEVICEID,
		"NVIDIA nForce MCP3 Networking Adapter"
	},
	{
		NFORCE_MCPNET4_DEVICEID,
		"NVIDIA nForce MCP5 Networking Adapter"
	},
	{
		NFORCE_MCPNET6_DEVICEID,
		"NVIDIA nForce MCP6 Networking Adapter"
	},
	{
		NFORCE_MCPNET7_DEVICEID,
		"NVIDIA nForce MCP7 Networking Adapter"
	},
	{
		NFORCE_MCPNET8_DEVICEID,
		"NVIDIA nForce MCP8 Networking Adapter"
	},
	{
		NFORCE_MCPNET9_DEVICEID,
		"NVIDIA nForce MCP9 Networking Adapter"
	},
	{
		NFORCE_MCPNET10_DEVICEID,
		"NVIDIA nForce MCP10 Networking Adapter"
	},
	{
		NFORCE_MCPNET11_DEVICEID,
		"NVIDIA nForce MCP11 Networking Adapter"
	},
	{
		0, NULL
	}
};
/* if_media interface */
static int nv_mediachange(struct ifnet *);
static void nv_mediastatus(struct ifnet *, struct ifmediareq *);
static void nv_update_stats(struct nv_softc *);

static int nv_ifmedia_upd(struct ifnet *);
static void nv_ifmedia_sts(struct ifnet *, struct ifmediareq *);
static int nv_miibus_readreg(struct device *, int, int);
static void nv_miibus_writereg(struct device *, int, int, int);
static void nv_miibus_statchg(struct device *);

/* ifnet interface */
static void nv_stop(struct ifnet *, int);
static int nv_init(struct ifnet *);
static int nv_ioctl(struct ifnet *, u_long, caddr_t);
static int nv_intr(void *);
static void nv_tick(void *);
static void nv_ifstart(struct ifnet * ifp);
static void nv_watchdog(struct ifnet *);
static void nv_setmulti(struct nv_softc *);

/* This is an internal function to keep nv_ioctl readable */
static int nv_lifaddr(struct ifnet *, u_long, struct ifaliasreq *);
static int nv_init_rings(struct nv_softc * sc);
static void nv_free_rings(struct nv_softc * sc);
static void nv_shutdown(void *);
/* ?? */
static int nv_pollinterval = 0;

static const struct nv_type *
nv_lookup(struct pci_attach_args * pa)
{
	const struct nv_type *nvt;

	if (PCI_VENDOR(pa->pa_id) != NVIDIA_VENDORID)
		return (NULL);

	for (nvt = nv_devs; nvt->nv_name != NULL; nvt++)
		if (PCI_PRODUCT(pa->pa_id) == nvt->nv_prodid)
			return nvt;

	return (NULL);
}
/* is device nVidia MCP Ethernet? */
static int
nv_match(struct device * self,
    struct cfdata * cfdata,
    void *arg)
{
	struct pci_attach_args *pa = arg;

	if (nv_lookup(pa) != NULL)
		return (1);

	return (0);
}

static void
nv_attach(struct device * parent, struct device * self, void *aux)
{
	struct nv_softc *sc = (struct nv_softc *) self;
	struct pci_attach_args *pa = aux;
	pci_chipset_tag_t pc = pa->pa_pc;
	pci_intr_handle_t ih;
	const char *intrstr = NULL;
	const struct nv_type *nvt;
	struct ifnet *ifp;
	bus_space_tag_t iot, memt;
	bus_space_handle_t ioh, memh;
	bus_addr_t dev_addr;
	bus_size_t dev_size;

	pcireg_t reg;
	int pmreg, ioh_valid, memh_valid;

	u_int8_t enaddr[ETHER_ADDR_LEN];
	unsigned long u;
	uint32_t ui;
	int i;
	int rtsegs;
	int rc;

	aprint_naive(": Ethernet Controller\n");
	nvt = nv_lookup(pa);
	if (nvt == NULL) {
		aprint_error("nv_attach impossible\n");
		return;
	}
	aprint_normal(": %s, rev %d\n", nvt->nv_name, PCI_REVISION(pa->pa_class));
	sc->pci_device_id = pa->pa_device;

	/* enable bus mastering */
	pci_conf_write(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG,
	    pci_conf_read(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG) |
	    PCI_COMMAND_MASTER_ENABLE);

	/*
         * allocate memory mapped address space for nv device
         */
	if ((rc = pci_mapreg_map(pa,
		    NV_RID,
		    PCI_MAPREG_TYPE_MEM | PCI_MAPREG_MEM_TYPE_32BIT,
		    BUS_SPACE_MAP_LINEAR,
		    &memt,
		    &memh,
		    &dev_addr,
		    &dev_size)) == 0) {
		sc->sc_st = memt;
		sc->sc_sh = memh;
		sc->sc_device_paddr = dev_addr;
		sc->sc_device_size = dev_size;
	} else {
		aprint_error("%s: unable to map device registers, error = %d\n",
		    sc->sc_dev.dv_xname,
		    rc);
		return;
	}

	/*
         * Map and establish our interrupt.
         */
	if (pci_intr_map(pa, &ih)) {
		aprint_error("%s: unable to map interrupt\n",
		    sc->sc_dev.dv_xname);
		return;
	}
	intrstr = pci_intr_string(pc, ih);
	sc->sc_ih = pci_intr_establish(pc, ih, IPL_NET, nv_intr, sc);
	if (sc->sc_ih == NULL) {
		aprint_error("%s: unable to establish interrupt",
		    sc->sc_dev.dv_xname);
		if (intrstr != NULL)
			aprint_normal(" at %s", intrstr);
		aprint_normal("\n");
		return;
	}
	aprint_normal("%s: interrupting at %s\n", sc->sc_dev.dv_xname, intrstr);

	/*
         * get my dma tag
         */
	sc->sc_dmat = pa->pa_dmat;

	/*
         * allocate dma memory and get the dma addresses for the nv_rx_desc's
         */
	{
		if ((rc = bus_dmamap_create(sc->sc_dmat,
			    sizeof(struct nv_rx_desc) * RX_RING_SIZE, 1,
			    sizeof(struct nv_rx_desc) * RX_RING_SIZE, 0,
			    BUS_DMA_NOWAIT,
			    &sc->sc_rmap)) != 0) {
			aprint_error("%s: Failed to create a dma map for nv_rx_desc memory - 0x%x\n",
			    sc->sc_dev.dv_xname, rc);
			return;
		}
		if ((rc = bus_dmamem_alloc(sc->sc_dmat,
			    sizeof(struct nv_rx_desc) * RX_RING_SIZE, PAGE_SIZE,
			    0,
			    &sc->sc_rxdesc_seg, 1,
			    &sc->sc_rxdesc_nsegs, BUS_DMA_NOWAIT | BUS_DMA_ALLOCNOW)) != 0) {
			aprint_error("%s: Failed to get nv_rx_desc dma memory - 0x%x\n",
			    sc->sc_dev.dv_xname, rc);
			return;
		}
		if ((rc = bus_dmamem_map(sc->sc_dmat,
			    &sc->sc_rxdesc_seg, 1,
			    sizeof(struct nv_rx_desc) * RX_RING_SIZE,
			    (caddr_t *) & sc->rx_desc,
			    BUS_DMA_NOWAIT | BUS_DMA_COHERENT)) != 0) {
			aprint_error("%s: Failed to map the nv_rx_desc memory - 0x%x\n",
			    sc->sc_dev.dv_xname, rc);
			return;
		}
		bzero(sc->rx_desc, sizeof(struct nv_rx_desc) * RX_RING_SIZE);
		if ((rc = bus_dmamap_load(sc->sc_dmat,
			    sc->sc_rmap,
			    (caddr_t *) sc->rx_desc,
			    sizeof(struct nv_rx_desc) * RX_RING_SIZE,
			    NULL,
			    BUS_DMA_NOWAIT)) != 0) {
			aprint_error("%s: Failed to load dma map of the nv_rx_desc memory - 0x%x\n",
			    sc->sc_dev.dv_xname, rc);
			return;
		}
#if 0
		aprint_normal("%s: nv_rx_desc: 0x%x - %d bytes\n",
		    sc->sc_dev.dv_xname,
		    sc->rx_desc,
		    sizeof(struct nv_rx_desc) * RX_RING_SIZE);
#endif
	}

	/*
         * allocate dma memory and get the dma addresses for the nv_tx_desc's
         */
	{
		if ((rc = bus_dmamap_create(sc->sc_dmat,
			    sizeof(struct nv_tx_desc) * TX_RING_SIZE, 1,
			    sizeof(struct nv_tx_desc) * TX_RING_SIZE, 0,
			    BUS_DMA_NOWAIT,
			    &sc->sc_tmap)) != 0) {
			aprint_error("%s: Failed to create a dma map for nv_tx_desc memory - 0x%x\n",
			    sc->sc_dev.dv_xname, rc);
			return;
		}
		if ((rc = bus_dmamem_alloc(sc->sc_dmat,
			    sizeof(struct nv_tx_desc) * TX_RING_SIZE, PAGE_SIZE,
			    0,
			    &sc->sc_txdesc_seg, 1,
			    &sc->sc_txdesc_nsegs, BUS_DMA_NOWAIT | BUS_DMA_ALLOCNOW)) != 0) {
			aprint_error("%s: Failed to get nv_tx_desc dma memory - 0x%x\n",
			    sc->sc_dev.dv_xname, rc);
			return;
		}
		if ((rc = bus_dmamem_map(sc->sc_dmat,
			    &sc->sc_txdesc_seg, 1,
			    sizeof(struct nv_tx_desc) * TX_RING_SIZE,
			    (caddr_t *) & sc->tx_desc,
			    BUS_DMA_NOWAIT | BUS_DMA_COHERENT)) != 0) {
			aprint_error("%s: Failed to map the nv_tx_desc memory - 0x%x\n",
			    sc->sc_dev.dv_xname, rc);
		}
		bzero(sc->tx_desc, sizeof(struct nv_tx_desc) * TX_RING_SIZE);
		if ((rc = bus_dmamap_load(sc->sc_dmat,
			    sc->sc_tmap,
			    (caddr_t *) sc->tx_desc,
			    sizeof(struct nv_tx_desc) * TX_RING_SIZE,
			    NULL,
			    BUS_DMA_NOWAIT)) != 0) {
			aprint_error("%s: Failed to load dma map of the nv_tx_desc memory - 0x%x\n",
			    sc->sc_dev.dv_xname, rc);
			return;
		}
#if 0
		aprint_normal("%s: nv_tx_desc: 0x%x - %d bytes\n",
		    sc->sc_dev.dv_xname,
		    sc->tx_desc,
		    sizeof(struct nv_tx_desc) * TX_RING_SIZE);
#endif
	}

	/*
         * initialize the NVidia TX/RX rings
         */
	nv_init_rings(sc);

	/* init nvidia api */
	nvidia_api_init(sc);

	sc->linkup = FALSE;
	sc->max_frame_size = ETHERMTU + ETHER_HDR_LEN + FCS_LEN;

	/* TODO - We don't support hardware offload yet */
	sc->hwmode = 1;
	sc->media = 0;

	if (nvidia_hardware_init(sc,
		2,
		RX_RING_SIZE,
		TX_RING_SIZE,
		6,
		nv_pollinterval) != 0) {
		aprint_error("nv_attach failed.\n");
		return;
	}
	bzero(&sc->adapterdata, sizeof(sc->adapterdata));
	sc->adapterdata.ulMediaIF = sc->media;
	sc->adapterdata.ulModeRegTxReadCompleteEnable = 1;
	sc->hwapi->pfnSetCommonData(sc->hwapi->pADCX, &sc->adapterdata);

	/* MAC is loaded backwards into h/w reg */
	sc->hwapi->pfnGetNodeAddress(sc->hwapi->pADCX, sc->original_mac_addr);
	for (i = 0; i < 6; i++) {
		enaddr[i] = sc->original_mac_addr[5 - i];
	}
	sc->hwapi->pfnSetNodeAddress(sc->hwapi->pADCX, enaddr);
	bcopy(enaddr, (char *) &sc->sc_macaddr, ETHER_ADDR_LEN);

	/* Display ethernet address ,... */
	aprint_normal("%s: Ethernet address %s\n",
	    sc->sc_dev.dv_xname,
	    ether_sprintf(enaddr));

	ifp = &sc->sc_ec.ec_if;

	/* setup the media and probe the MII */
	sc->sc_mii.mii_ifp = ifp;
	sc->sc_mii.mii_readreg = nv_miibus_readreg;
	sc->sc_mii.mii_writereg = nv_miibus_writereg;
	sc->sc_mii.mii_statchg = nv_miibus_statchg;
	ifmedia_init(&sc->sc_mii.mii_media, IFM_IMASK, nv_ifmedia_upd, nv_ifmedia_sts);
	mii_attach(&sc->sc_dev, &sc->sc_mii, 0xffffffff, MII_PHY_ANY,
	    MII_OFFSET_ANY, /* miiflags */ 0);
	if (LIST_FIRST(&sc->sc_mii.mii_phys) == NULL) {
		ifmedia_add(&sc->sc_mii.mii_media, IFM_ETHER | IFM_NONE, 0, NULL);
		ifmedia_set(&sc->sc_mii.mii_media, IFM_ETHER | IFM_NONE);
	} else
		ifmedia_set(&sc->sc_mii.mii_media, IFM_ETHER | IFM_AUTO);

	/* setup the interface */
	strcpy(ifp->if_xname, sc->sc_dev.dv_xname);
	ifp->if_softc = sc;
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	ifp->if_ioctl = nv_ioctl;
	ifp->if_start = nv_ifstart;
	ifp->if_stop = nv_stop;
	ifp->if_init = nv_init;
	ifp->if_watchdog = nv_watchdog;
	ifp->if_timer = 0;
	ifp->if_mtu = ETHERMTU;
	ifp->if_baudrate = IF_Mbps(100);
	IFQ_SET_READY(&ifp->if_snd);
	ifp->if_snd.ifq_maxlen = TX_RING_SIZE - 1;

	sc->sc_ec.ec_capabilities = ETHERCAP_VLAN_MTU | ETHERCAP_JUMBO_MTU;

	/* finally attach the interface */
	if_attach(ifp);
	ether_ifattach(ifp, enaddr);

	/* setup shutdown hook ... */
	sc->sc_sdhook = shutdownhook_establish(nv_shutdown, sc);
	if (sc->sc_sdhook == NULL) {
		aprint_error("%s: WARNING: unable to establish shutdown hook\n",
		    sc->sc_dev.dv_xname);
		return;
	}
	/* mark as ATTACHED ... */
	sc->sc_flags |= NV_ATTACHED;

	DEBUGOUT(NV_DEBUG_INIT, "nv: nv_attach - exit\n");

}
/*
 * When detaching, we do the inverse of what is done in the attach
 * routine, in reversed order.
 */
static int
nv_detach(struct device * self, int flags)
{
	struct nv_softc *sc = (struct nv_softc *) self;
	struct ifnet *ifp = &sc->sc_ec.ec_if;

	if ((sc->sc_flags & NV_ATTACHED) == 0)
		return 0;

	/* remove ticker */
	callout_stop(&sc->sc_stat_ch);

	/* detach all PHYs */
	mii_detach(&sc->sc_mii, MII_PHY_ANY, MII_OFFSET_ANY);

	/* delete all remaining media */
	ifmedia_delete_instance(&sc->sc_mii.mii_media, IFM_INST_ANY);

	ether_ifdetach(ifp);
	if_detach(ifp);

	/* remove tx/rx desc rings */
	nv_free_rings(sc);

	/* restore the original reversed mac address */
	sc->hwapi->pfnSetNodeAddress(sc->hwapi->pADCX, sc->original_mac_addr);

	/* remove dma mappings */
	bus_dmamap_destroy(sc->sc_dmat, sc->sc_rmap);
	bus_dmamem_free(sc->sc_dmat, &sc->sc_rxdesc_seg, sc->sc_rxdesc_nsegs);
	bus_dmamap_destroy(sc->sc_dmat, sc->sc_tmap);
	bus_dmamem_free(sc->sc_dmat, &sc->sc_txdesc_seg, sc->sc_txdesc_nsegs);

	/* unmap pci regs ... */
	bus_space_unmap(sc->sc_st, sc->sc_sh, sc->sc_device_size);

	/* report and remember we've detached */
	if (!(flags & DETACH_QUIET))
		aprint_normal("%s detached\n",
		    self->dv_xname);

	sc->sc_flags &= ~NV_ATTACHED;
	return (0);
}


/* Main loop for sending packets from OS to interface */
static void
nv_ifstart(struct ifnet * ifp)
{

	struct nv_softc *sc = ifp->if_softc;
	struct nv_map_buffer *buf;
	struct mbuf *m0, *m;
	struct nv_tx_desc *desc;
	ADAPTER_WRITE_DATA txdata;
	int error, i;

	DEBUGOUT(NV_DEBUG_RUNNING, "nv: nv_ifstart - entry\n");

	/* If link is down/busy or queue is empty do nothing */
	if (ifp->if_flags & IFF_OACTIVE || ifp->if_snd.ifq_head == NULL) {
		return;
	}
	/* Transmit queued packets until sent or TX ring is full */
	while (sc->pending_txs < TX_RING_SIZE) {
		desc = sc->tx_desc + sc->cur_tx;
		buf = &desc->buf;

		/* Get next packet to send. */
		IF_DEQUEUE(&ifp->if_snd, m0);

		/* If nothing to send, return. */
		if (m0 == NULL)
			return;
#if 0
		aprint_normal("%s: nv_ifstart mbuf = 0x%x, m_len = %d to send.\n",
		    sc->sc_dev.dv_xname,
		    m0,
		    m0->m_len);
#endif

		/* Map MBUF for DMA access */
		error = bus_dmamap_load_mbuf(sc->sc_dmat,
		    buf->map,
		    m0,
		    BUS_DMA_WRITE | BUS_DMA_NOWAIT);
		if (error && error != EFBIG) {
			aprint_error("%s: nv_ifstart - bus_dmamap_load_mbuf failed, error = %d\n",
			    sc->sc_dev.dv_xname,
			    error);

			m_freem(m0);
			sc->tx_errors++;
			continue;
		}
		/*
	         * Packet has too many fragments - defrag into new mbuf
	         * cluster
	         */
		if (error) {
#if 0
			aprint_error("%s: Packet has too many fragments, attempting to coalesce ...\n",
			    sc->sc_dev.dv_xname);
#endif
			MGETHDR(m, M_DONTWAIT, MT_DATA);
			if (m == NULL) {
				m_freem(m0);
				sc->tx_errors++;
				aprint_error("%s: unable to allocate Tx mbuf\n",
				    sc->sc_dev.dv_xname);
				continue;
			}
			if (m0->m_pkthdr.len > MHLEN) {
				MCLGET(m, M_DONTWAIT);
				if ((m->m_flags & M_EXT) == 0) {
					aprint_error("%s: unable to allocate Tx "
					    "cluster\n", sc->sc_dev.dv_xname);
					m_freem(m);
					sc->tx_errors++;
					continue;
				}
			}
			m_copydata(m0, 0, m0->m_pkthdr.len, mtod(m, caddr_t));
			m->m_pkthdr.len = m->m_len = m0->m_pkthdr.len;

			error = bus_dmamap_load_mbuf(sc->sc_dmat,
			    buf->map,
			    m,
			    BUS_DMA_WRITE | BUS_DMA_NOWAIT);
			if (error) {
				m_freem(m);
				sc->tx_errors++;
				continue;
			}
			m_freem(m0);
			m0 = m;
		}
		/* need to get mapping info on mbuf for sending to device */
		for (i = 0; i < buf->map->dm_nsegs; i++) {
			desc->frags[i] = buf->map->dm_segs[i];
		}

		desc->numfrags = buf->map->dm_nsegs;
#if 0
		aprint_normal("%s: mbuf has %d segs\n",
		    sc->sc_dev.dv_xname,
		    buf->map->dm_nsegs);
#endif
		/* Do sync on DMA bounce buffer */
		bus_dmamap_sync(sc->sc_dmat,
		    buf->map,
		    0, buf->map->dm_mapsize,
		    BUS_DMASYNC_PREWRITE);
		buf->mbuf = m0;
		txdata.ulNumberOfElements = desc->numfrags;
		txdata.pvID = (PVOID) desc;

		/* Put fragments into API element list */
		txdata.ulTotalLength = buf->mbuf->m_len;
		for (i = 0; i < desc->numfrags; i++) {
			txdata.sElement[i].ulLength = (ulong) desc->frags[i].ds_len;
			txdata.sElement[i].pPhysical = (PVOID) desc->frags[i].ds_addr;
		}

		/* Send packet to Nvidia API for transmission */
		error = sc->hwapi->pfnWrite(sc->hwapi->pADCX, &txdata);
		switch (error) {
		case ADAPTERERR_NONE:
			/* Packet was queued in API TX queue successfully */
			sc->pending_txs++;
			sc->cur_tx = (sc->cur_tx + 1) % TX_RING_SIZE;
			break;

		case ADAPTERERR_TRANSMIT_QUEUE_FULL:
			/* The API TX queue is full - requeue the packet */
			ifp->if_flags |= IFF_OACTIVE;
			bus_dmamap_unload(sc->sc_dmat, buf->map);
			IF_PREPEND(&ifp->if_snd, buf->mbuf);
			buf->mbuf = NULL;
			return;

		default:
			/* The API failed to queue/send the packet so dump it */
			bus_dmamap_unload(sc->sc_dmat, buf->map);
			m_freem(buf->mbuf);
			buf->mbuf = NULL;
			sc->tx_errors++;
			return;
		}
		/* Set watchdog timer. */
		ifp->if_timer = 8;

		/* Copy packet to BPF tap */
		BPF_MTAP(ifp, m0);
	}
	ifp->if_flags |= IFF_OACTIVE;

	DEBUGOUT(NV_DEBUG_RUNNING, "nv: nv_ifstart - exit\n");
}

static int
nv_ioctl(struct ifnet * ifp, u_long cmd, caddr_t data)
{
	struct nv_softc *sc = (struct nv_softc *) ifp->if_softc;
	struct ifreq *ifr = (struct ifreq *) data;
	int s, error;

	s = splnet();

	switch (cmd) {
	case SIOCSIFMEDIA:
	case SIOCGIFMEDIA:
		error = ifmedia_ioctl(ifp, ifr, &sc->sc_mii.mii_media, cmd);
		break;
	case SIOCSIFPHYADDR:
		error = nv_lifaddr(ifp, cmd, (struct ifaliasreq *) data);
		break;
	default:
		error = ether_ioctl(ifp, cmd, data);
		if (error == ENETRESET) {
			/*
			 * Multicast list has changed; set the hardware filter
			 * accordingly.  Update our idea of the current media;
			 * nv_setmulti() needs to know what it is.
			 */
			if (ifp->if_flags & IFF_RUNNING) {
				mii_pollstat(&sc->sc_mii);
				nv_setmulti(sc);
			}
			error = 0;
		}
		break;
	}

	splx(s);

	return (error);
}
/*
 * Helper function to set Ethernet address.  This shouldn't be done there,
 * and should actually be available to all Ethernet drivers, real or not.
 */
static int
nv_lifaddr(struct ifnet * ifp, u_long cmd, struct ifaliasreq * ifra)
{
	struct sockaddr *sa = (struct sockaddr *) & ifra->ifra_addr;

	if (sa->sa_family != AF_LINK)
		return (EINVAL);

	memcpy(LLADDR(ifp->if_sadl), sa->sa_data, ETHER_ADDR_LEN);

	return (0);
}
/*
 * _init() would typically be called when an interface goes up,
 * meaning it should configure itself into the state in which it
 * can send packets.
 */
static int
nv_init(struct ifnet * ifp)
{
	struct nv_softc *sc = ifp->if_softc;
	int error;
	NV_LOCK(sc);

	DEBUGOUT(NV_DEBUG_INIT, "nv: nv_init - entry (%d)\n", sc->linkup);

	/* Do nothing if already running */
	if (ifp->if_flags & IFF_RUNNING)
		goto fail;

	nv_stop(ifp, 0);

	DEBUGOUT(NV_DEBUG_INIT, "nv: do pfnInit\n");
	/* Setup Hardware interface and allocate memory structures */
	error = sc->hwapi->pfnInit(sc->hwapi->pADCX,
	    0,			/* force speed */
	    0,			/* force full duplex */
	    0,			/* force mode */
	    0,			/* force async mode */
	    &sc->linkup);
	if (error) {
		aprint_error("%s: failed to start NVIDIA Hardware interface\n",
		    sc->sc_dev.dv_xname);
		goto fail;
	}
	/* Set the MAC address */
	sc->hwapi->pfnSetNodeAddress(sc->hwapi->pADCX, sc->sc_macaddr);
	sc->hwapi->pfnEnableInterrupts(sc->hwapi->pADCX);
	sc->hwapi->pfnStart(sc->hwapi->pADCX);

	/* Setup multicast filter */
	nv_setmulti(sc);
	nv_ifmedia_upd(ifp);

	/* Update interface parameters */
	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;

	/* set the timer */
	callout_setfunc(&sc->sc_stat_ch, nv_tick, sc);
	callout_schedule(&sc->sc_stat_ch, hz);

	DEBUGOUT(NV_DEBUG_INIT, "nv: nv_init - exit\n");

fail:
	NV_UNLOCK(sc);

	return (0);
}
/*
 * _stop() is called when an interface goes down.  It is our
 * responsability to validate that state by clearing the
 * IFF_RUNNING flag.
 */
static void
nv_stop(struct ifnet * ifp, int disable)
{
	struct nv_softc *sc = ifp->if_softc;

	NV_LOCK(sc);

	DEBUGOUT(NV_DEBUG_RUNNING, "nv: nv_stop - entry\n");

	ifp->if_timer = 0;

	/* Cancel tick timer */
	callout_stop(&sc->sc_stat_ch);

	/* Stop hardware activity */
	sc->hwapi->pfnDisableInterrupts(sc->hwapi->pADCX);
	sc->hwapi->pfnStop(sc->hwapi->pADCX, 0);

	DEBUGOUT(NV_DEBUG_DEINIT, "nv: do pfnDeinit\n");
	/* Shutdown interface and deallocate memory buffers */
	if (sc->hwapi->pfnDeinit)
		sc->hwapi->pfnDeinit(sc->hwapi->pADCX, 0);

	sc->linkup = 0;
	sc->cur_rx = 0;
	sc->pending_rxs = 0;

	ifp->if_flags &= ~(IFF_RUNNING | IFF_OACTIVE);

	DEBUGOUT(NV_DEBUG_RUNNING, "nv: nv_stop - exit\n");

	NV_UNLOCK(sc);

	return;
}
/* Shutdown interface for unload/reboot */
static void
nv_shutdown(void *p)
{
	struct nv_softc *sc = p;

	DEBUGOUT(NV_DEBUG_DEINIT, "nv: nv_shutdown\n");

	/* Stop hardware activity */
	nv_stop(&sc->sc_ec.ec_if, 0);
}
/* Allocate TX ring buffers */
static int
nv_init_rings(struct nv_softc * sc)
{
	int error, i;

	NV_LOCK(sc);

	DEBUGOUT(NV_DEBUG_INIT, "nv: nv_init_rings - entry\n");

	sc->cur_rx = sc->cur_tx = sc->pending_rxs = sc->pending_txs = 0;

	/* Initialize RX ring */
	for (i = 0; i < RX_RING_SIZE; i++) {
		struct nv_rx_desc *desc = sc->rx_desc + i;
		struct nv_map_buffer *buf = &desc->buf;

		MGETHDR(buf->mbuf, M_DONTWAIT, MT_DATA);
		MCLGET(buf->mbuf, M_DONTWAIT);
		if (buf->mbuf == NULL) {
			aprint_error("%s: Initialize RX Ring - couldn't allocate an mbuf.\n",
			    sc->sc_dev.dv_xname);
			nv_free_rings(sc);
			error = ENOBUFS;
			goto fail;
		}
		buf->mbuf->m_len = buf->mbuf->m_pkthdr.len = MCLBYTES;
		m_adj(buf->mbuf, ETHER_ALIGN);

		error = bus_dmamap_create(sc->sc_dmat,
		    buf->mbuf->m_len, 1,
		    buf->mbuf->m_len, 0,
		    BUS_DMA_NOWAIT,
		    &buf->map);
		if (error) {
			aprint_error("%s: Initialize RX Ring - couldn't create dma map\n",
			    sc->sc_dev.dv_xname);
			nv_free_rings(sc);
			goto fail;
		}
		error = bus_dmamap_load_mbuf(sc->sc_dmat,
		    buf->map,
		    buf->mbuf,
		    BUS_DMA_NOWAIT);
		if (error) {
			aprint_error("%s: Initialize RX Ring - couldn't dma map mbuf\n",
			    sc->sc_dev.dv_xname);
			nv_free_rings(sc);
			goto fail;
		}
		bus_dmamap_sync(sc->sc_dmat, buf->map, 0, buf->mbuf->m_len, BUS_DMASYNC_PREREAD);

		desc->buflength = buf->mbuf->m_len;
		desc->vaddr = mtod(buf->mbuf, caddr_t);
		desc->paddr = vtophys(desc->vaddr);
	}
	bus_dmamap_sync(sc->sc_dmat,
	    sc->sc_rmap,
	    0, sizeof(struct nv_rx_desc) * RX_RING_SIZE,
	    BUS_DMASYNC_PREREAD | BUS_DMASYNC_PREWRITE);

	/* Initialize TX ring */
	for (i = 0; i < TX_RING_SIZE; i++) {
		struct nv_tx_desc *desc = sc->tx_desc + i;
		struct nv_map_buffer *buf = &desc->buf;

		buf->mbuf = NULL;

		error = bus_dmamap_create(sc->sc_dmat,
		    MCLBYTES, 1,
		    MCLBYTES, 0,
		    BUS_DMA_NOWAIT,
		    &buf->map);
		if (error) {
			aprint_error("%s: Initialize TX Ring - couldn't create dma map\n",
			    sc->sc_dev.dv_xname);
			nv_free_rings(sc);
			goto fail;
		}
	}
	bus_dmamap_sync(sc->sc_dmat,
	    sc->sc_tmap,
	    0, sizeof(struct nv_tx_desc) * TX_RING_SIZE,
	    BUS_DMASYNC_PREREAD | BUS_DMASYNC_PREWRITE);

	DEBUGOUT(NV_DEBUG_INIT, "nv: nv_init_rings - exit\n");

fail:
	NV_UNLOCK(sc);

	return (error);
}
/* Free the TX ring buffers */
static void
nv_free_rings(struct nv_softc * sc)
{
	int i;

	NV_LOCK(sc);

	DEBUGOUT(NV_DEBUG_DEINIT, "nv: nv_free_rings - entry\n");

	for (i = 0; i < RX_RING_SIZE; i++) {
		struct nv_rx_desc *desc = sc->rx_desc + i;
		struct nv_map_buffer *buf = &desc->buf;

		if (buf->mbuf) {
			bus_dmamap_unload(sc->sc_dmat, buf->map);
			bus_dmamap_destroy(sc->sc_dmat, buf->map);
			m_freem(buf->mbuf);
		}
		buf->mbuf = NULL;
	}

	for (i = 0; i < TX_RING_SIZE; i++) {
		struct nv_tx_desc *desc = sc->tx_desc + i;
		struct nv_map_buffer *buf = &desc->buf;

		if (buf->mbuf) {
			bus_dmamap_unload(sc->sc_dmat, buf->map);
			bus_dmamap_destroy(sc->sc_dmat, buf->map);
			m_freem(buf->mbuf);
		}
		buf->mbuf = NULL;
	}

	DEBUGOUT(NV_DEBUG_DEINIT, "nv: nv_free_rings - exit\n");

	NV_UNLOCK(sc);
}


/* Interrupt service routine */
static int
nv_intr(void *arg)
{
	struct nv_softc *sc = arg;
	struct ifnet *ifp = &sc->sc_if;

	if (!ifp->if_flags & IFF_UP) {
		nv_stop(ifp, 0);
		return;
	}
	/* Handle interrupt event */
	if (sc->hwapi->pfnQueryInterrupt(sc->hwapi->pADCX)) {
		sc->hwapi->pfnHandleInterrupt(sc->hwapi->pADCX);
		sc->hwapi->pfnEnableInterrupts(sc->hwapi->pADCX);
	}
	if (ifp->if_snd.ifq_head != NULL)
		nv_ifstart(ifp);

	/* If no pending packets we don't need a timeout */
	if (sc->pending_txs == 0)
		sc->sc_if.if_timer = 0;

	return;
}

/* Setup multicast filters */
static void
nv_setmulti(struct nv_softc * sc)
{
	struct ethercom *ec = &sc->sc_ec;
	struct ifnet *ifp = &sc->sc_ec.ec_if;
	struct ether_multi *enm;
	struct ether_multistep step;
	PACKET_FILTER hwfilter;
	int i;
	u_int8_t oraddr[6];
	u_int8_t andaddr[6];

	NV_LOCK(sc);

	DEBUGOUT(NV_DEBUG_RUNNING, "nv: nv_setmulti - entry\n");

	ifp = &sc->sc_if;

	/* Initialize nvidia filter structure */
	hwfilter.ulFilterFlags = 0;
	for (i = 0; i < 6; i++) {
		hwfilter.acMulticastAddress[i] = 0;
		hwfilter.acMulticastMask[i] = 0;
	}

	/* do we want all packets? */
	if (ifp->if_flags & (IFF_PROMISC | IFF_ALLMULTI)) {
allmulti:
		/* Yep, so accept all packets */
		hwfilter.ulFilterFlags |= ACCEPT_ALL_PACKETS;
		sc->hwapi->pfnSetPacketFilter(sc->hwapi->pADCX, &hwfilter);
		NV_UNLOCK(sc);
		return;
	}

	/* Nope so setup the multicast filters */
	ETHER_FIRST_MULTI(step, ec, enm);
	while (enm != NULL) {
	        if (memcmp(enm->enm_addrlo, enm->enm_addrhi, ETHER_ADDR_LEN) != 0) {
			/*
			 * We must listen to a range of multicast addresses.
			 * For now, just accept all multicasts, rather than
			 * trying to set only those filter bits needed to match
			 * the range.  (At this time, the only use of address
			 * ranges is for IP multicast routing, for which the
			 * range is big enough to require all bits set.)
			 */
		        goto allmulti;
		}

		u_char *addrp;

		addrp = enm->enm_addrlo;
		for (i = 0; i < 6; i++) {
			u_int8_t mcaddr = addrp[i];
			andaddr[i] &= mcaddr;
			oraddr[i] |= mcaddr;
		}
		ETHER_NEXT_MULTI(step, enm);
	}

	/* create the nvidia filter */
	for (i = 0; i < 6; i++) {
		hwfilter.acMulticastAddress[i] = andaddr[i] & oraddr[i];
		hwfilter.acMulticastMask[i] = andaddr[i] | (~oraddr[i]);
	}

	/* now send the filter to the device (via the NVIDIA API) */
	sc->hwapi->pfnSetPacketFilter(sc->hwapi->pADCX, &hwfilter);

	NV_UNLOCK(sc);

	DEBUGOUT(NV_DEBUG_RUNNING, "nv: nv_setmulti - exit\n");

	return;
}


/* Change the current media/mediaopts */
static int
nv_ifmedia_upd(struct ifnet * ifp)
{
	struct nv_softc *sc = ifp->if_softc;
	struct mii_data *mii = &sc->sc_mii;
	struct ifmedia *ifm = &mii->mii_media;
	int media = ifm->ifm_cur->ifm_media;
	u_int32_t miicfg;
	struct mii_softc *miisc;
	int cfg;
	DEBUGOUT(NV_DEBUG_MII, "nv: nv_ifmedia_upd\n");

	if (!(ifp->if_flags & IFF_UP))
		return 0;

	if (mii->mii_instance) {
		struct mii_softc *miisc;
		for (miisc = LIST_FIRST(&mii->mii_phys); miisc != NULL;
		    miisc = LIST_NEXT(miisc, mii_list)) {
			mii_phy_reset(miisc);
		}
	}
	mii_mediachg(mii);
	return (0);
}
/* Update current miibus PHY status of media */
static void
nv_ifmedia_sts(struct ifnet * ifp, struct ifmediareq * ifmr)
{
	struct nv_softc *sc = ifp->if_softc;

	DEBUGOUT(NV_DEBUG_MII, "nv: nv_ifmedia_sts\n");

	mii_pollstat(&sc->sc_mii);

	ifmr->ifm_active = sc->sc_mii.mii_media_active;
	ifmr->ifm_status = sc->sc_mii.mii_media_status;

	return;
}
/* miibus tick timer - maintain link status */
static void
nv_tick(void *xsc)
{
	struct nv_softc *sc = xsc;
	struct mii_data *mii = &sc->sc_mii;
	struct ifnet *ifp = &sc->sc_if;

	NV_LOCK(sc);

	nv_update_stats(sc);
	mii_tick(mii);

	if (mii->mii_media_status & IFM_ACTIVE &&
	    IFM_SUBTYPE(mii->mii_media_active) != IFM_NONE) {
		if (ifp->if_snd.ifq_head != NULL)
			nv_ifstart(ifp);
	}
	callout_schedule(&sc->sc_stat_ch, hz);

	NV_UNLOCK(sc);

	return;
}


/* Update ifnet data structure with collected interface stats from API */
static void
nv_update_stats(struct nv_softc * sc)
{
	struct ifnet *ifp = &sc->sc_if;
	ADAPTER_STATS stats;

	NV_LOCK(sc);

	if (sc->hwapi) {
		sc->hwapi->pfnGetStatistics(sc->hwapi->pADCX, &stats);

		ifp->if_ipackets = stats.ulSuccessfulReceptions;
		ifp->if_ierrors = stats.ulMissedFrames +
		    stats.ulFailedReceptions +
		    stats.ulCRCErrors +
		    stats.ulFramingErrors +
		    stats.ulOverFlowErrors;

		ifp->if_opackets = stats.ulSuccessfulTransmissions;
		ifp->if_oerrors = sc->tx_errors +
		    stats.ulFailedTransmissions +
		    stats.ulRetryErrors +
		    stats.ulUnderflowErrors +
		    stats.ulLossOfCarrierErrors +
		    stats.ulLateCollisionErrors;

		ifp->if_collisions = stats.ulLateCollisionErrors;
	}
	NV_UNLOCK(sc);

	return;
}

static void
nv_miibus_statchg(struct device * dev)
{
}
/* miibus Read PHY register wrapper - calls Nvidia API entry point */
static int
nv_miibus_readreg(struct device * dev, int phy, int reg)
{
	struct nv_softc *sc = (struct nv_softc *) dev;
	ULONG data;

	ADAPTER_ReadPhy(sc->hwapi->pADCX, phy, reg, &data);

	return (data);
}
/* miibus Write PHY register wrapper - calls Nvidia API entry point */
static void
nv_miibus_writereg(struct device * dev, int phy, int reg, int data)
{
	struct nv_softc *sc = (struct nv_softc *) dev;

	//DEBUGOUT(NV_DEBUG_MII, "nv: nv_miibus_writereg - entry\n");

	ADAPTER_WritePhy(sc->hwapi->pADCX, phy, reg, (ulong) data);

	//DEBUGOUT(NV_DEBUG_MII, "nv: nv_miibus_writereg - exit\n");

	return;
}


/* Watchdog timer to prevent PHY lockups */
static void
nv_watchdog(struct ifnet * ifp)
{
	struct nv_softc *sc = ifp->if_softc;

	aprint_error("%s: device timeout (%d)\n", sc->sc_dev.dv_xname, sc->pending_txs);
	sc->tx_errors++;

	nv_stop(ifp, 0);
	ifp->if_flags &= ~IFF_RUNNING;
	nv_init(ifp);

	if (ifp->if_snd.ifq_head != NULL)
		nv_ifstart(ifp);

	return;
}

static char *
format_txdesc(struct nv_tx_desc * desc)
{
	static char buffer[4096] = {0};
	sprintf(buffer,
	    "   next: 0x%x\n"
	    "   buf:{\n"
	    "         mbuf: 0x%x\n"
	    "         map:  dmap\n"
	    "   }\n"
	    "   buflength: %d\n"
	    "   numfrags:  %d\n",
	    desc->next,
	    desc->buf.mbuf,
	    desc->buflength,
	    desc->numfrags);
	return buffer;
}
